/********************************************************************
 Software License Agreement:

 The software supplied herewith by Microchip Technology Incorporated
 (the "Company") for its PIC(R) Microcontroller is intended and
 supplied to you, the Company's customer, for use solely and
 exclusively on Microchip PIC Microcontroller products. The
 software is owned by the Company and/or its supplier, and is
 protected under applicable copyright laws. All rights are reserved.
 Any use in violation of the foregoing restrictions may subject the
 user to criminal sanctions under applicable laws, as well as to
 civil liability for the breach of the terms and conditions of this
 license.

 THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION. NO WARRANTIES,
 WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
 TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
 IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
 CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 *******************************************************************/

/** INCLUDES *******************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "system.h"
#include "system_config.h"

#include "usb/usb.h"
#include "usb/usb_device_hid.h"

#include "app_device_custom_hid.h"
#include "debug.h"
#include "patterns.h"
#include "ascii_table.h"

char debugString[40];

uint8_t p_intensity=0,table_type=0,cycle_count =0,pattern_speed=0;
uint8_t run = 1;
uint8_t demo_mode = 0;
uint8_t user_id = 0;
uint8_t plockout[4];
uint8_t p_count=0;
uint8_t p_table=0;
uint8_t fader_cycle=0;
uint8_t demo_loops;
uint8_t old_button = 0;
uint8_t led_data_save[5];

void update_pattern(void);
void update_pattern(void);
void next_pattern(void);
void back_pattern(void);
uint8_t get_next_pattern_byte(void);
uint8_t up_down( uint8_t, uint8_t);
uint8_t pattern_done(void);
void add_and_shift( uint8_t);
void handle_push_button(void);

const uint8_t rom_settings[0x10] @ 0x1f00;

MAIN_RETURN main(void)
{
    int delay=0;
//  start DDF

//    ANSELA = 0x00;
    ANSELC = 0x00;
    LATA = 0x00;
    LATC = 0x00;
    TRISA  = 0x00; // trying to get A4 and A5 to work
    TRISC  = 0x00;

    OSCTUNE = 0;
    OSCCON = 0x7C;   // PLL enabled, 3x, 16MHz internal osc, SCS external
    OSCCONbits.SPLLMULT = 1;   // 1=3x, 0=4x
    ACTCON = 0x90;   // Clock recovery on, Clock Recovery enabled; SOF packet
    OPTION_REGbits.nWPUEN = 0;  //enable weak pullups.
    WPUAbits.WPUA3 = 1; // weak pullup on A3
    WPUAbits.WPUA4 = 0;
    WPUAbits.WPUA5 = 0;
    
    demo_mode = rom_settings[0];
    user_msg_size = rom_settings[1];
    user_id = rom_settings[2];
    plockout[0] = rom_settings[3];
    plockout[1] = rom_settings[4];
    plockout[2] = rom_settings[5];
    plockout[3] = rom_settings[6];
    
    // holding button on powerup turns on demo mode.
    if (!sw0) {demo_mode = 0x80;}

//  end DDF
    SYSTEM_Initialize(SYSTEM_STATE_USB_START);

    USBDeviceInit();
    USBDeviceAttach();

 	// Initialise the debug log functions
    debugInitialise();

 	sprintf(debugString, "2DKits: 7x8 matrix %d %d", demo_mode, user_msg_size);
	debugOut(debugString);

//   debugOut(debugString);

        //  Enable Timer0
        //           +------------------ 0 Weak Pullup enabled
        //           |+----------------- 1 edge (don'r care)
        //           ||+---------------- 0 instuction clock
        //           |||+--------------- 1 edge (don't care)
        //           ||||+-------------- 1 PSA assigned
        //           |||||+++----------- 1 prescale 1:2
        OPTION_REG = 0b01000000;
        INTCONbits.TMR0IE = 1; // enable timer0 interupt
       
    while(1)
    {
        SYSTEM_Tasks();

        #if defined(USB_POLLING)
            // Interrupt or polling method.  If using polling, must call
            // this function periodically.  This function will take care
            // of processing and responding to SETUP transactions
            // (such as during the enumeration process when you first
            // plug in).  USB hosts require that USB devices should accept
            // and process SETUP packets in a timely fashion.  Therefore,
            // when using polling, this function should be called
            // regularly (such as once every 1.8ms or faster** [see
            // inline code comments in usb_device.c for explanation when
            // "or faster" applies])  In most cases, the USBDeviceTasks()
            // function does not take very long to execute (ex: <100
            // instruction cycles) before it returns.
            USBDeviceTasks();
        #endif

        /* If the USB device isn't configured yet, we can't really do anything
         * else since we don't have a host to talk to.  So jump back to the
         * top of the while loop. */
  //      if( USBGetDeviceState() < CONFIGURED_STATE )
  //      {
            /* Jump back to the top of the while loop. */
  //          continue;
  //      }

        /* If we are currently suspended, then we need to see if we need to
         * issue a remote wakeup.  In either case, we shouldn't process any
         * keyboard commands since we aren't currently communicating to the host
         * thus just continue back to the start of the while loop. */
  //      if( USBIsDeviceSuspended() == true )
  //      {
            /* Jump back to the top of the while loop. */
  //          continue;
  //      }

        //Application specific tasks
        if( USBGetDeviceState() == CONFIGURED_STATE ) {
           APP_DeviceCustomHIDTasks();
        }
        // Note: Other application specific actions can be placed here
         if ((wait_timer == 0) && (run == 1)) {
            wait_timer = 1;
            delay++;
#ifndef DEBUG
            if ((!sw0) || old_button) {
                handle_push_button();
            } else {
#endif
               if (delay > pattern_speed*16) {
                  delay = 0;
                  update_pattern();
               }
#ifndef DEBUG
            }
#endif
         }




    }//end while
}//end main

void update_pattern() {

   uint8_t i, j, a, b, x, y, temp_in;
   static uint8_t temp_red[8], char_in, step, delay2;

// pull extra info bytes off the front of the pattern;

   if (p_count == 0) {
      table_type = get_next_pattern_byte( );
      pattern_speed = get_next_pattern_byte( );
      step = 0;
   }

   switch (table_type) {
   case 1:
      for (i=0;i<8;i++) {
         temp_in = get_next_pattern_byte();
         LEDS[i][0] = (temp_in & 0x80) ? (0x07) : (0x00);
         LEDS[i][1] = (temp_in & 0x40) ? (0x07) : (0x00);
         LEDS[i][2] = (temp_in & 0x20) ? (0x07) : (0x00);
         LEDS[i][3] = (temp_in & 0x10) ? (0x07) : (0x00);
         LEDS[i][4] = (temp_in & 0x08) ? (0x07) : (0x00);
         LEDS[i][5] = (temp_in & 0x04) ? (0x07) : (0x00);
         LEDS[i][6] = (temp_in & 0x02) ? (0x07) : (0x00);
         LEDS[i][7] = (temp_in & 0x01) ? (0x07) : (0x00);
      }
      break;
   case 8:
   case 9:
      if (fader_cycle == 0) {
         if (table_type == 9) {srand((unsigned)TMR0);}
         for (i=0;i<8;i++) {
            if (table_type == 8) {
            temp_red[i] = get_next_pattern_byte( );
            } else if (table_type == 9) {
            temp_red[i] = rand();
            }
         }
         cycle_count = get_next_pattern_byte( );
      }
      for (i=0;i<8;i++) {
         LEDS[i][0] = up_down((temp_red[i] & 0x80),LEDS[i][0]);
         LEDS[i][1] = up_down((temp_red[i] & 0x40),LEDS[i][1]);
         LEDS[i][2] = up_down((temp_red[i] & 0x20),LEDS[i][2]);
         LEDS[i][3] = up_down((temp_red[i] & 0x10),LEDS[i][3]);
         LEDS[i][4] = up_down((temp_red[i] & 0x08),LEDS[i][4]);
         LEDS[i][5] = up_down((temp_red[i] & 0x04),LEDS[i][5]);
         LEDS[i][6] = up_down((temp_red[i] & 0x02),LEDS[i][6]);
         LEDS[i][7] = up_down((temp_red[i] & 0x01),LEDS[i][7]);
      }
      fader_cycle = (fader_cycle < cycle_count) ? (fader_cycle+1) : (0);
//       	sprintf(debugString, "A %d %d", fader_cycle, cycle_count);
//	debugOut(debugString);
      break;
case 4:
      if (fader_cycle == 0) {
         for (i=0;i<8;i++) {
            temp_red[i] = 0;
            for (j=0;j<8;j++) {
                switch (( rand()>>4) % 2) {
                case 0:
                   break;
                case 1:
                   temp_red[i] = temp_red[i] + (1<<j);
                   break;
                }
            }
         }
         i = rand(); //without this you get the same
         cycle_count = get_next_pattern_byte( );
      }
      for (i=0;i<8;i++) {
         LEDS[i][1] = up_down((temp_red[i] & 0x80),LEDS[i][1]);
         LEDS[i][2] = up_down((temp_red[i] & 0x40),LEDS[i][2]);
         LEDS[i][3] = up_down((temp_red[i] & 0x20),LEDS[i][3]);
         LEDS[i][4] = up_down((temp_red[i] & 0x10),LEDS[i][4]);
         LEDS[i][5] = up_down((temp_red[i] & 0x08),LEDS[i][5]);
         LEDS[i][6] = up_down((temp_red[i] & 0x04),LEDS[i][6]);
         LEDS[i][7] = up_down((temp_red[i] & 0x02),LEDS[i][7]);
      }
      fader_cycle = (fader_cycle < cycle_count) ? (fader_cycle+1) : (0);
      break;
case 3:
    delay2++;
    if (delay2 > 6) {
       if (step==0) {
          char_in = get_next_pattern_byte();
          while (char_in < 20) {
             char_in = get_next_pattern_byte();
          }
          char_in = char_in - 0x20;
          add_and_shift( 0x00);
          step = 1;
       } else {
          add_and_shift( chardata[char_in].ch[step-1]);
          if (step>=chardata[char_in].sl) { step=0; }
          else { step++; }
       }
       delay2 = 0;
     }
     break;
case 6:
     add_and_shift( get_next_pattern_byte());
     break;
case 7:
    do {
       char_in = get_next_pattern_byte();
       y = (char_in & 0b00011100)>>2;
       x = (char_in & 0b11100000)>>5;
       a = (char_in & 0b00000010)>>1;
       b = (char_in & 0b00000001);
       if (y == 0) {
           switch ((x<<1)+a) {
             case 0:  // clear all
             case 1:  // set all
                 for (i=0;i<8;i++) {
                     for (j=0;j<8;j++) {
                         LEDS[i][j] = a;
                     }
                 }
                 break;
             case 2: // shift down fill 0
             case 3: // shift down fill 1
                 for (i=0;i<8;i++) {
                    for (j=1;j<7;j++) {
                       LEDS[i][j] =   LEDS[i][j+1];
                    }
                 }
                 for (i=0;i<8;i++) {
                     LEDS[i][7] = a;
                 }
                 break;
             case 6: // shift right fill 0
             case 7: // shift right fill 1
                 for (i=7;i>0;i--) {
                    for (j=1;j<8;j++) {
                       LEDS[i][j] =   LEDS[i-1][j];
                    }
                 }
                 for (i=1;i<8;i++) {
                     LEDS[0][i] = a;
                 }
                 break;
          }
       } else {
         LEDS[x][y] = a;
       }
    } while (b==0);
    break;
   }

//   sprintf(debugString, "B %d %d %d", p_count, p_table, pattern_size[p_table]);
//   debugOut(debugString);
   if (pattern_done()) {
      p_count = 0;
      demo_loops = demo_loops - 1;
      if ((demo_loops == 0) && (demo_mode == 0x80)) {
         next_pattern();
      }
   }
}

uint8_t pattern_done()
{
    if ((p_table == 0) && (patterns[USER_DEFINED_PATTERN][0] == 3)) {
      return(p_count >=user_msg_size);
   } else {
      return(p_count >= pattern_size[p_table]);
   }
}

void add_and_shift( uint8_t in) {

   uint8_t i,j;
   for (i=0;i<7;i++) {
       for (j=1;j<=7;j++) {
          LEDS[i][j] =   LEDS[i+1][j];
       }
   }

   LEDS[7][7]   = (in   & 0x80) ? (0x07) : (0x00);
   LEDS[7][6]   = (in   & 0x40) ? (0x07) : (0x00);
   LEDS[7][5]   = (in   & 0x20) ? (0x07) : (0x00);
   LEDS[7][4]   = (in   & 0x10) ? (0x07) : (0x00);
   LEDS[7][3]   = (in   & 0x08) ? (0x07) : (0x00);
   LEDS[7][2]   = (in   & 0x04) ? (0x07) : (0x00);
   LEDS[7][1]   = (in   & 0x02) ? (0x07) : (0x00);
}

uint8_t get_next_pattern_byte( )
{
   if ((p_table == 0) && (patterns[USER_DEFINED_PATTERN][0] == 3)) {
      return(patterns[USER_DEFINED_PATTERN][p_count++]);
   } else {
      return(patterns[p_table][p_count++]);
   }
}

void handle_push_button() {

unsigned char i, temp;
int delay_count;
   if (old_button == 0) {
      for (i=0;i<4;i++) {
         led_data_save[i] = LEDS[0][7-i];      // save the 4 LED states
         LEDS[0][7-i] = 0x00;

      }
      led_data_save[4] = LEDS[0][7];
      LEDS[0][7] = 0x00;

   }

   temp = p_table+1;

   LEDS[1][7] = (demo_mode == 0x80) ? (0xff) : (0x00);  // display the table number
   LEDS[0][7] = (temp & 0x01) ? (0xff) : (0x00);  // display the table number
   LEDS[0][6] = (temp & 0x02) ? (0xff) : (0x00);
   LEDS[0][5] = (temp & 0x04) ? (0xff) : (0x00);
   LEDS[0][4] = (temp & 0x08) ? (0xff) : (0x00);

   old_button++;

#define BUTTON_DELAY 250

#ifndef DEBUG
   if (!sw0) {
       delay_count = 3600;
       do {
          delay_count--;
       } while(delay_count);

      if (old_button > BUTTON_DELAY) {
         next_pattern();
         old_button = 1;
      }
   } else {
      for (i=0;i++;i<4) {
         LEDS[0][7-i] = led_data_save[i];  // no button, restore the 4 LEDs state.
      }
      LEDS[1][7] = led_data_save[4];  // no button, restore the 4 LEDs state.
      old_button = 0;
      sprintf(debugString, "pattern %d %d", p_table, demo_mode);
	  debugOut(debugString);
   }
#endif
}



void next_pattern(void) {
      p_table = (p_table >= table_count-1) ? (0):(p_table+1);
      p_count = 0;
      demo_loops = pattern_demo_loops[p_table];
    sprintf(debugString, "Up to pattern %d %d", p_table, demo_loops);
	  debugOut(debugString);
}

void back_pattern(void) {
     p_table = (p_table < 1) ? (table_count-1):(p_table-1);
     p_count = 0;
     demo_loops = pattern_demo_loops[p_table];
   sprintf(debugString, "Down to pattern %d %d", p_table, demo_loops);
	 debugOut(debugString);
}

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

uint8_t up_down( uint8_t change, uint8_t data) {

  if (change) {
      return(MIN(8,data+1));
  } else {
      return(MAX(0,(signed char) data-1));
  }
}




bool USER_USB_CALLBACK_EVENT_HANDLER(USB_EVENT event, void *pdata, uint16_t size)
{
    switch((int)event)
    {
        case EVENT_TRANSFER:
            break;

        case EVENT_SOF:
            break;

        case EVENT_SUSPEND:
            break;

        case EVENT_RESUME:
            break;

        case EVENT_CONFIGURED:
            /* When the device is configured, we can (re)initialize the demo
             * code. */
            APP_DeviceCustomHIDInitialize();
            break;

        case EVENT_SET_DESCRIPTOR:
            break;

        case EVENT_EP0_REQUEST:
            /* We have received a non-standard USB request.  The HID driver
             * needs to check to see if the request was for it. */
            USBCheckHIDRequest();
            break;

        case EVENT_BUS_ERROR:
            break;

        case EVENT_TRANSFER_TERMINATED:
            break;

        default:
            break;
    }
    return true;
}

/*******************************************************************************
 End of File
*/

